/* Copyright (c) 2013, Colfax International. All Right Reserved.
Redistribution or commercial usage without a written permission 
from Colfax International is prohibited.
Contact information can be found at http://colfax-intl.com/     */


#include <stdio.h>
#include <mkl.h>
#include <math.h>
#include <omp.h>

#define HARMONIC 4.0
#define TWOPI 6.2831853071795864769

void init_data(MKL_Complex16* data, const long fft_size,  const long num_fft) {
  const double k = HARMONIC*TWOPI/(double)(fft_size);
#pragma omp parallel for
  for(int i = 0; i < num_fft; i++) {
    for(int j = 0; j < fft_size; j++) {
      data[i*fft_size+j].real = cos(k*j);
      data[i*fft_size+j].imag = sin(k*j);
    }
  }
}


int main(int argc, char* argv[]) {

  if(argc < 3) {
    printf("Usage: %s [fft_size] [num_fft]\n", argv[0]);
    return 1;
  }

  const long fft_size = atol(argv[1]);
  const long num_fft = atol(argv[2]);
  const int nTrials=10;
  const int skipTrials=2;

  
  MKL_Complex16* data = (MKL_Complex16*) malloc(sizeof(MKL_Complex16)*fft_size*num_fft);
  
  DFTI_DESCRIPTOR_HANDLE handle;
  DftiCreateDescriptor(&handle, DFTI_DOUBLE, DFTI_COMPLEX, 1, (MKL_LONG)fft_size);
  DftiSetValue(handle, DFTI_NUMBER_OF_TRANSFORMS, num_fft);
  DftiSetValue(handle, DFTI_INPUT_DISTANCE, fft_size);
  DftiSetValue(handle, DFTI_OUTPUT_DISTANCE, fft_size);
  DftiSetValue(handle, DFTI_PLACEMENT, DFTI_INPLACE);
  DftiCommitDescriptor(handle);

  init_data(data, fft_size, num_fft);
  printf("\nPerforming %ld 1D FFT of size %ld\n", num_fft, fft_size);
  printf("MKL     threads: %d\nOpenMP affinity: %s\n", mkl_get_max_threads(), getenv("KMP_AFFINITY"));
  printf("\033[1m%5s %10s %8s\033[0m\n", "Trial", "Time, s", "GFLOP/s");
  double perf=0.0, dperf=0.0;
  const double HztoPerf = 5.0*1e-9*double(fft_size)*log2(double(fft_size))*double(num_fft);
  for (int t=1; t<=nTrials; t++){
    const double t0 = omp_get_wtime();
    DftiComputeForward(handle, data);    
    const double t1 = omp_get_wtime();

    if (t == 0) {
      double err = 0.0;
      for(int i = 0; i < num_fft; i++) {
	err += (data[i*fft_size+(size_t)HARMONIC].real - (double)fft_size)*(data[i*fft_size+(size_t)HARMONIC].real - (double)fft_size);
      }
      err = sqrt(err);
      if(err > 1e-6) {
	printf("Something is wrong...\nerr=%f\nExiting...\n", err);
	return 2;
      }
    }

    if (t > skipTrials) {
      perf += HztoPerf/(t1-t0);
      dperf += HztoPerf*HztoPerf/((t1-t0)*(t1-t0));
    }
    
    printf("%5d %10.3e %8.1f %s\n", 
	   t, (t1-t0), HztoPerf/(t1-t0), (t<=skipTrials?"*":""));
    fflush(0);
  }
  perf/=(double)(nTrials-skipTrials); 
  dperf=sqrt(dperf/(double)(nTrials-skipTrials)-perf*perf);
  printf("-----------------------------------------------------\n");
  printf("\033[1m%s %4s \033[42m%10.1f +- %.1f GFLOP/s\033[0m\n",
	 "Average performance:", "", perf, dperf);
  printf("-----------------------------------------------------\n");
  printf("* - warm-up, not included in average\n\n");

}
