/* Copyright (c) 2013, Colfax International. All Right Reserved.
Redistribution or commercial usage without a written permission 
from Colfax International is prohibited.
Contact information can be found at http://colfax-intl.com/     */


#include <stdio.h>
#include <mkl.h>
#include <math.h>
#include <omp.h>

#define HARMONIC 4.0
#define TWOPI 6.2831853071795864769

void init_data(MKL_Complex16* data, const long fft_size_x, const long fft_size_y,  const long num_fft) {
  const double kx = HARMONIC*TWOPI/(double)(fft_size_x);
  const double ky = HARMONIC*TWOPI/(double)(fft_size_y);
  const long mat_size = fft_size_x*fft_size_y;   
#pragma omp parallel for
  for(int i = 0; i < num_fft; i++) {
    for(int j = 0; j < fft_size_y; j++) {
      for(int k = 0; k < fft_size_x; k++) {
	data[i*mat_size+fft_size_x*j+k].real = cos(ky*j)*cos(kx*k);
	data[i*mat_size+fft_size_x*j+k].imag = -sin(ky*j)*sin(kx*k);
      }
    }
  }
}

int main(int argc, char* argv[]) {

  if(argc < 4) {
    printf("Usage: %s [fft_size_x] [fft_size_y] [num_fft]\n", argv[0]);
    return 1;
  }

  const long fft_size_x = atol(argv[1]);
  const long fft_size_y = atol(argv[2]);
  MKL_LONG DIM[2] = {fft_size_y, fft_size_x};
  const long num_fft = atol(argv[3]);
  const int nTrials=10;
  const int skipTrials=2;
  const int num_threads=omp_get_max_threads();

  DFTI_DESCRIPTOR_HANDLE handle[num_threads];

  for(int i = 0; i < num_threads; i++) {
    DftiCreateDescriptor(&handle[i], DFTI_DOUBLE, DFTI_COMPLEX, 2, DIM);
    DftiSetValue(handle[i], DFTI_INPUT_DISTANCE, fft_size_x*fft_size_y);
    DftiSetValue(handle[i], DFTI_INPUT_DISTANCE, fft_size_x*fft_size_y);
    DftiSetValue(handle[i], DFTI_OUTPUT_DISTANCE, fft_size_x*fft_size_y);
    DftiSetValue(handle[i], DFTI_PLACEMENT, DFTI_INPLACE);
    DftiCommitDescriptor(handle[i]);
  }


  MKL_Complex16* data = (MKL_Complex16*)mkl_malloc(sizeof(MKL_Complex16)*fft_size_x*fft_size_y*num_fft, 64);
  init_data(data, fft_size_x, fft_size_y, num_fft);
  printf("\nPerforming %ld 2D FFT on %ld-by-%ld matrices\n", num_fft, fft_size_x, fft_size_y); fflush(0);

  printf(" Affinity: %s\n\n", getenv("KMP_AFFINITY"));
  printf("\033[1m%5s %10s %8s\033[0m\n", "Trial", "Time, s", "GFLOP/s");


  double perf=0.0, dperf=0.0;
  const double HztoPerf = 5.0*1e-9*double(fft_size_x*fft_size_y)*log2(double(fft_size_x*fft_size_y))*double(num_fft);
  for (int t=1; t<=nTrials; t++){
    const double t0 = omp_get_wtime();
#pragma omp parallel
    {
      const int t_id = omp_get_thread_num();
#pragma omp for
      for(long i = 0; i < num_fft; i++) {
	DftiComputeForward(handle[t_id], &data[i*fft_size_x*fft_size_y]);
      }
    }
    const double t1 = omp_get_wtime();

    if (t > skipTrials) {
      perf += HztoPerf/(t1-t0);
      dperf += HztoPerf*HztoPerf/((t1-t0)*(t1-t0));
    }
    
    printf("%5d %10.3e %8.1f %s\n", 
	   t, (t1-t0), HztoPerf/(t1-t0), (t<=skipTrials?"*":""));
    fflush(0);
  }
  perf/=(double)(nTrials-skipTrials); 
  dperf=sqrt(dperf/(double)(nTrials-skipTrials)-perf*perf);
  printf("-----------------------------------------------------\n");
  printf("\033[1m%s %4s \033[42m%10.1f +- %.1f GFLOP/s\033[0m\n",
	 "Average performance:", "", perf, dperf);
  printf("-----------------------------------------------------\n");
  printf("* - warm-up, not included in average\n\n");
}
