#include <cmath>
#include <cstdio>
#include <omp.h>
#include <cassert>
#include <mkl.h>

void init_matrix(double* matrix, int nMatrices, int n) {
#pragma omp parallel for schedule(guided)
  for (int m = 0; m < nMatrices; m++) {
    for (int i = 0; i < n; i++) {
        double sum = 0.0f;
        for (int j = 0; j < n; j++) {
	  matrix[i*n+j+m*n*n] = (double)(i*n+j);
	  sum += matrix[i*n+j+m*n*n];
        }
        sum -= matrix[i*n+i+m*n*n];
        matrix[i*n+i+m*n*n] = 2.0*sum;
    }
  }

}


int main(const int argc, const char** argv) {

  if(argc < 2) {
    printf("Usage: %s [size] [nMatrices]", argv[0]);
    return 1;
  }

  // Problem size and other parameters
  const lapack_int  n=atoi(argv[1]);
  const int nMatrices=atoi(argv[2]);
  
  const double HztoPerf = 1e-9*2.0/3.0*((double)n*(double)n*(double)n)*(double)nMatrices;

  printf("LU decomposition of %d matrices of size %dx%d (%.3f GB) on %s...\n\n", 
	 nMatrices, n, n, 8.0*(double)n*(double)n*(double)nMatrices*1.0e-9,
#ifndef __MIC__
	 "CPU"
#else
	 "MIC"
#endif
	 ); fflush(stdout);

  printf("OpenMP teams: %d\n", omp_get_max_threads());
#pragma omp parallel
  {
#pragma omp master
    {
      printf("OpenMP threads per team: %d\n", omp_get_max_threads());
    }
  }

  // Align on 2 MB to get a fresh new page
  double* matrix = (double*) _mm_malloc(sizeof(double)*(size_t)n*(size_t)n*(size_t)nMatrices, (1<<21));

  // Initialize matrices
  init_matrix(matrix, nMatrices, n);  

  // Perform benchmark
  double rate = 0, dRate = 0; // Benchmarking data
  const int nTrials = 10;
  const int skipTrials = 3; // First step is warm-up on Xeon Phi coprocessor
  printf("\033[1m%5s %10s %8s\033[0m\n", "Trial", "Time, s", "GFLOP/s"); fflush(stdout);
  for (int trial = 1; trial <= nTrials; trial++) {
    lapack_int ipiv[n];
    // Reference calculation to verify results
    init_matrix(matrix, nMatrices, n);
    const double tStart = omp_get_wtime(); // Start timing
#pragma omp parallel for schedule(static)
    for (int m = 0; m < nMatrices; m++) {
      LAPACKE_dgetrf(LAPACK_ROW_MAJOR, n, n, &matrix[n*n*m], n, &ipiv[0]);
    }
    const double tEnd = omp_get_wtime(); // End timing

    if (trial > skipTrials) { // Collect statistics
      rate  += HztoPerf/(tEnd - tStart); 
      dRate += HztoPerf*HztoPerf/((tEnd - tStart)*(tEnd-tStart)); 
    }

    printf("%5d %10.3e %8.2f %s\n", 
	   trial, (tEnd-tStart), HztoPerf/(tEnd-tStart), (trial<=skipTrials?"*":""));
    fflush(stdout);
  }
  rate/=(double)(nTrials-skipTrials); 
  dRate=sqrt(dRate/(double)(nTrials-skipTrials)-rate*rate);
  printf("-----------------------------------------------------\n");
  printf("\033[1m%s %4s \033[42m%10.2f +- %.2f GFLOP/s\033[0m\n",
	 "Average performance:", "", rate, dRate);
  printf("-----------------------------------------------------\n");
  printf("* - warm-up, not included in average\n\n");

  _mm_free(matrix);

}
